// use std::sync::atomic::{AtomicUsize, Ordering};
use std::sync::Arc;
use tokio::sync::Mutex;
use tokio_cron_scheduler::{JobScheduler, Job};
pub type BxDynResult<T> = std::result::Result<T, Box<dyn std::error::Error>>;
use app_env::{
  AppStore,
  appenv::AppEnv,
  appinfo::AppInfo,
  appdata::AppData,
  config::{Config}
};
use clds::monitor::defs::MonitorRules;
use connectors::defs::{AppDataConn}; 
use app_auth::AuthStore;
use reject_filters::{handle_rejection};
use anyhow::{anyhow,Result};
use warp::{
  // http::{StatusCode},
  http::{method::Method, HeaderMap},
  Filter,
};
use crate::defs::{DataDBs,CollsData,load_cloud_env};
use clds::clouds::defs::{
  Cloud,
};
use clds::clouds::on_clouds::{make_cloud_cache,run_clouds_check};
use reqenv::ReqEnv;

// static WEBSERVER: AtomicUsize = AtomicUsize::new(0);
const PKG_VERSION: &'static str = env!("CARGO_PKG_VERSION");
// const PKG_VERSION: Option<&'static str> = option_env!("CARGO_PKG_VERSION");
const PKG_NAME: &'static str = env!("CARGO_PKG_NAME");
const PKG_AUTHORS: &'static str = env!("CARGO_PKG_AUTHORS");
const PKG_DESCRIPTION: &'static str = env!("CARGO_PKG_DESCRIPTION");

pub mod defs;
pub mod graphql;
pub mod filters;
pub mod handlers;

async fn create_auth_store(app_env: &AppEnv,verbose: isize) -> AuthStore {
	let config = app_env.get_curr_websrvr_config();
	let model_path = config.st_auth_model_path();
  let policy_path = config.st_auth_policy_path();
  AuthStore::new(&config,AuthStore::create_enforcer(model_path,policy_path).await,verbose)
}
async fn up_web_server(webpos: usize) -> Result<()> {
	let debug = envmnt::get_isize("DEBUG",0);
	let verbose = envmnt::get_isize("WEB_SERVER_VERBOSE", 0);
	let mut app_env = AppEnv::default();
	app_env.curr_web=webpos;
	if verbose > 0 {
		println!("Web services: init {} ___________ ", chrono::Utc::now().timestamp());
	}
	app_env.info = AppInfo::new(
		"ZTerton",
		format!("web: {}",&webpos),
		format!("version: {}",PKG_VERSION),
		format!("authors: {}",PKG_AUTHORS),
		format!("{}",PKG_DESCRIPTION),
	);
	zterton::init_app(&mut app_env,verbose).await.unwrap_or_else(|e| 
		panic!("Error loadding app environment {}",e)
	);
	let config = app_env.get_curr_websrvr_config();
	// let webserver_status = WEBSERVER.load(Ordering::Relaxed);
	let zterton_env = envmnt::get_or(format!("ZTERTON_{}",&config.name).as_str(), "UNKNOWN");
	// if webserver_status != 0 {
	if zterton_env != "UNKNOWN" {
		if verbose  > 0 {
			println!("ZTerton web services at {}",&zterton_env);
		}
		return Ok(());
	}
	// WEBSERVER.store(1,Ordering::Relaxed);
  // TODO pass root file-name frmt from AppEnv Config 
	if verbose > 0 {
		println!("Loading webserver: {} ({})",&config.name,&app_env.curr_web);
	}	
  let (app, socket) = zterton::start_web(&mut app_env).await;
	if verbose > 0 {
 	 println!("Load app store ...");
	}
  let app_store = AppStore::new(AppData::new(app_env.to_owned(),verbose));
  // As static casbin
	if verbose > 0 {
  	println!("Load auth store ...");
	}
	let auth_store = create_auth_store(&app_env,verbose).await;
	if verbose > 0 {
  	println!("Load data store ...");
	}
	let app_data_conn = AppDataConn::new("Zterton".to_string(),app_env.config.datastores_settings.to_owned(),"").await;
  let data_dbs = DataDBs {
    colls: CollsData::new(app_env.to_owned(),verbose),
    app: app_store.to_owned(),
    auth: auth_store.to_owned(),
		conns: Arc::new(Mutex::from(app_data_conn)),
  };
	if verbose > 0 {
 	 println!("Load web filters ...");
	}

  // let us get some static boxes from config values: 
  let log_name = app_env.config.st_log_name();
  // Path for static files 
  let html_path = config.st_html_path();
  // If not graphQL comment/remove next line 
  let gql_path = config.st_gql_req_path();
  // If not graphiQL comment/remove next line Interface GiQL
  let giql_path = config.st_giql_req_path();

  let origins: Vec<&str> = config.allow_origin.iter().map(AsRef::as_ref).collect();
  let cors = warp::cors()
    //.allow_any_origin()
    .allow_origins(origins)
    //.allow_origins(vec![app_env.config.allow_origin.as_str(), "https://localhost:8000"])
    .allow_credentials(true)
    .allow_header("content-type")
    .allow_header("Authorization")
    .allow_methods(&[Method::GET, Method::POST, Method::DELETE]);

  let auth_api = 
  // Auth routes for login & logout  REQUIRED
  app_auth_filters::auth(app_store.clone(),auth_store.clone(),cors.clone()); // .with(cors.clone());
  
  let gqli_api = 
  // If not graphiQL comment/remove next line Interface GiQL MUST BEFORE graphql post with schema
  // app_api.to_owned()
  graphql::graphiql(gql_path, giql_path, data_dbs.clone()).await;
  if giql_path.len() > 0 && verbose > 0 {
    println!(
     "GraphiQL url: {}://{}:{}/{}",
     &app.protocol, &app.host, &app.port, &giql_path
    );
  } 
	let mut cloud = Cloud::default();
	load_cloud_env(&mut cloud).await;
  // app_api.to_owned()
  // If not graphQL comment/remove next line 
  let gql_api=graphql::graphql(gql_path, data_dbs.clone(),cors.clone()).await; //.with(cors.clone());

  // // Add ALL ENTITIES to work with here
  let kloud_api = filters::CollFilters::new("kloud")
		.filters_config(data_dbs.clone(),cloud.clone(),cors.clone());

   let file_api = app_file_filters::files(app_store.clone(),auth_store.clone()).with(cors.clone());
  // Path for static files, better to be LAST
   let fs_api = warp::fs::dir(html_path).with(warp::compression::gzip());
  // Recover and handle errors 
  let app_api = auth_api
    .or(gqli_api).or(gql_api)
    .or(kloud_api)
    .or(file_api)
    .or(fs_api)
    .recover(move | error: warp::Rejection| handle_rejection(error, app_store.clone()))
    .boxed();
  // Wrap routes with log to get info
  let routes =  app_api.with(warp::log(log_name));
  //let routes = app_api.with(cors).with(warp::log(log_name));
  println!(
    "Starting http server: {}://{}:{}",
    &app.protocol, &app.host, &app.port
  );
	envmnt::set("ZTERTON", format!("{}:{}",&app.host,&app.port));
	if debug > 0 {
 	  println!("Web services: done {} __________ ",chrono::Utc::now().timestamp());
	}
	if app.protocol.clone().as_str() == "http" {
		warp::serve(routes.to_owned())
			.run(socket)
			.await;
	} else {
		warp::serve(routes)
			.tls()
			.cert_path(format!("{}/ssl/{}", &config.resources_path, "cert.pem"))
			.key_path(format!("{}/ssl/{}", &config.resources_path, "key.pem"))
			.run(socket)
			.await;
	}
	Ok(())
}
fn get_args() -> (String,String) {
	let args: Vec<String> = std::env::args().collect();
	let mut arg_cfg_path = String::from("");
	let mut arg_env_path = String::from("");
	args.iter().enumerate().for_each(|(idx,arg)| {
		if arg == "-c" { 
			arg_cfg_path=args[idx+1].to_owned();
		} else if arg == "-e" { 
			arg_env_path=args[idx+1].to_owned();
		}
  });
	(arg_cfg_path,arg_env_path)
}
async fn get_app_env(arg_cfg_path: String,verbose: isize) -> Result<(Cloud,AppEnv)> {
	let mut cloud = Cloud::default();
	load_cloud_env(&mut cloud).await;	
	let mut app_env = AppEnv::default();
	let config_content = Config::load_file_content(verbose,&arg_cfg_path);
  if ! config_content.contains("run_mode") {
		Err(anyhow!("Run mode not found in config {}", &arg_cfg_path))
	} else {
		app_env.config = Config::new(config_content,verbose);
		Ok((cloud,app_env))
	}
}
async fn set_reqenv(app_env: &AppEnv,verbose: isize) -> ReqEnv {
	let app_store =	AppStore::new(AppData::new(app_env.to_owned(),verbose));
	let auth_store = create_auth_store(&app_env,verbose).await;
	let mut headers = HeaderMap::new();
  headers.insert(http::header::HOST, "localhost".parse().unwrap());
	ReqEnv::new( 
		app_store, 
		auth_store,
		headers,
	  Method::GET,
		"/config", "config", "kloud"
	)
}
pub async fn run_cache_clouds() -> Result<()> {
	let (arg_cfg_path,arg_env_path) = get_args();
	let now = chrono::Utc::now().timestamp();
	let verbose = envmnt::get_isize("DEBUG", 0);
	if verbose > 0 {
  	println!("Cache service on Clouds: run {} __________ {}  {} ",&now,&arg_cfg_path,&arg_env_path);
	}
	let (cloud, app_env) = match get_app_env(arg_cfg_path,verbose).await {
		Ok((c,e)) => (c,e),
		Err(e) => {
      println!("Cache service on Clouds: done {} __________ ",&now);
			return Err(e);
		},
	};
	let reqenv = set_reqenv(&app_env,verbose).await;
	let res = make_cloud_cache(&reqenv,&cloud).await;
	if verbose > 0 {
  	println!("Cache service on Clouds: done {} __________ ",&now);
	}
	res
}
pub async fn run_check_clouds() -> Result<()> {
	let (arg_cfg_path,arg_env_path) = get_args();
	let now = chrono::Utc::now().timestamp();
	let verbose = envmnt::get_isize("DEBUG", 0);
	if verbose > 0 {
 	  println!("Check Cloud services: run {} __________ {} {} ",&now,&arg_cfg_path,&arg_env_path);
	}
	let (cloud, app_env) = match get_app_env(arg_cfg_path,verbose).await {
		Ok((c,e)) => (c,e),
		Err(e) => {
  	  println!("Check Cloud service: done {} __________ ",&now);
			return Err(e);
		},
	};
	let reqenv = set_reqenv(&app_env,verbose).await;
	let res = run_clouds_check(&reqenv,&cloud).await;
	if verbose > 0 {
 	   println!("Check Cloud service: done {} __________ ",&now);
	}
	res
}
pub async fn run_clouds_monitor() -> Result<()> {
	let (arg_cfg_path,arg_env_path) = get_args();
	let now = chrono::Utc::now().timestamp();
	let verbose = envmnt::get_isize("DEBUG", 0);
	if verbose > 0 {
  	println!("Monitor Cloud: run {} __________ {}  {} ",&now,arg_cfg_path,&arg_env_path);
	}
	let (cloud, app_env) = match get_app_env(arg_cfg_path,verbose).await {
		Ok((c,e)) => (c,e),
		Err(e) => {
  		println!("Monitor Cloud done {} __________ ",&now);
			return Err(e);
		},
	};	
	let monitor_rules = MonitorRules::load(
		&app_env.config.monitor_rules_path,
		&app_env.config.monitor_rules_file,
		&app_env.config.monitor_rules_format
	);
	if monitor_rules.rules.len() == 0 {
		eprintln!("No monitor rules found");
		return Ok(());
	} 
	let res = monitor_rules.run(cloud,app_env).await;
	if verbose > 0 {
 	 println!("Monitor Cloud done {} __________ ",&now);
	}
	res
}
#[tokio::main] 
pub async fn main() -> BxDynResult<()> {  //std::io::Result<()> {  
	let args: Vec<String> = std::env::args().collect();
	if args.len() > 1 {
		match args[1].as_str() {
			"-h" | "--help" => 
	  		println!("{} USAGE: -c config-toml -e env.file",PKG_NAME),
			"-v" | "--version" => {
	  		println!("{} version: {}",PKG_NAME,PKG_VERSION);
				return Ok(());
				},
			_ => println!("{}",PKG_NAME),
		}
	}
	let debug=envmnt::get_isize("DEBUG",0);
	let mut sched = JobScheduler::new();
	let mut arg_cfg_path = String::from("");
	let mut arg_env_path = String::from("");
	args.iter().enumerate().for_each(|(idx,arg)| {
		if arg == "-c" { 
			arg_cfg_path=args[idx+1].to_owned();
		} else if arg == "-e" { 
			arg_env_path=args[idx+1].to_owned();
		}
  });
	if !arg_env_path.is_empty() {
		let env_path = std::path::Path::new(&arg_env_path);
		dotenv::from_path(env_path)?;
	}
	pretty_env_logger::init();
	let config_content = Config::load_file_content(debug, &arg_cfg_path);
	if !config_content.contains("run_mode") {
		panic!("Error no run_mode found or config path incomplete");
	}
	let config = Config::new(config_content,debug);
	let app_data_conn = AppDataConn::new("Zterton".to_string(),config.datastores_settings.to_owned(),"").await;
	if !app_data_conn.check_connections(config.datastores_settings.to_owned()).await {
		panic!("Error checking app data store connections");
	} 
	if config.run_schedtasks {
		for it in &config.schedtasks {
			if ! it.on_start {
				continue;
			}
			match it.name.as_str() {
				"monitor" => tokio::spawn(async {run_clouds_monitor().await}),
				"check" => tokio::spawn(async {run_check_clouds().await}),
				"cache" => tokio::spawn(async {run_cache_clouds().await}),
				_ => { 
					eprintln!("Error task {} not defined",&it.name);
					continue;
				},
			};
		}
	}
	if config.run_websrvrs {
		for (pos,it) in config.websrvrs.iter().enumerate() {
			if debug > 1 {
				println!("{} -> {}",it.name,pos);
			}
			tokio::spawn(async move {up_web_server(pos).await});
			tokio::time::sleep(tokio::time::Duration::from_secs(3)).await;
		}
	}
	if config.run_schedtasks {
		for it in config.schedtasks {
			if it.schedule.is_empty() {
				eprintln!("Task {} no schedule defined",&it.name);
				continue;
			}
			let res = match it.name.as_str() {
				"monitor" => 
					sched.add(Job::new(&it.schedule.to_owned(), move |uuid, _l| {
						if debug > 0 {
							println!("Schedule {} {}: {}",&it.name,&it.schedule,uuid);
						}
						tokio::spawn(async {run_clouds_monitor().await});
					})?),
				"check" => 
					sched.add(Job::new(&it.schedule.to_owned(), move |uuid, _l| {
						if debug > 0 {
							println!("Schedule {} {}: {}",&it.name,&it.schedule,uuid);
						}
						tokio::spawn(async {run_check_clouds().await});
					})?),
				"cache" => 
					sched.add(Job::new(&it.schedule.to_owned(), move |uuid, _l| {
						if debug > 0 {
							println!("Schedule {} {}: {}",&it.name,&it.schedule,uuid);
						}
						tokio::spawn(async {run_cache_clouds().await});
					})?),
				_ => { 
					eprintln!("Error task {} not defined",&it.name);
					continue;
				},
			};
			match res {
				Ok(_) => { continue; },
				Err(e) => { 
					eprintln!("Error scheduling task {}",e);
					continue;
				},
			}
		};
		let _= sched.start().await;
	}
	Ok(())
}
